import Foundation

// n * m の大きさの迷路を作成する
/*
 (1) スタート - まず全てを壁にする。そしてスタート位置から掘ってゆく。
 (2) 自分のいる周りを調べて、四方がすでに通路ならば帰る
 (3) 進むことのできる方角の配列をシャッフルしてループ
 (4) ループの順番で掘り進み、
 (5) 自分で掘った場所が通路で囲まれているかもしれないので、自分のいる座標周りを調べ、四方が通路だったら継続
 (6) そうで無いならば、自分自身の関数を呼び出して再帰
 */


// 列挙型 => 掘れる壁、掘れない壁、通路、スタート位置、ゴール位置
public enum MazeWallType
{
    case Digable    // 掘れる
    case NotDigable // 掘れない
    case Digged     // 掘ったから通路になった
    case Start      // スタート位置。掘れない
    case Goal       // ゴール位置。掘れるが、掘ったらゴールがなくなる....
}

// 迷路の中の座標を表す。NSPointやCGPointと違い整数値しか持たない。
public struct MazePos : Hashable, Equatable
{
    let x : Int
    let y : Int
    
    
    public init(_ x: Int, _ y: Int)
    {
        self.x = x
        self.y = y
    }
    
    public static func == (lhs: MazePos, rhs: MazePos) -> Bool
    {
        return lhs.x == rhs.x && lhs.y == rhs.y
    }

    public static func != (lhs: MazePos, rhs: MazePos) -> Bool
    {
        return lhs.x != rhs.x || lhs.y != rhs.y
    }

    // MazePos同士の加算を+記号で行う。
    public static func + (lhs: MazePos, rhs: MazePos) -> MazePos
    {
        return MazePos(lhs.x + rhs.x, lhs.y + rhs.y)
    }
    
    public var hashValue: Int 
    { 
        return "\(x), \(y)".hashValue
    }
}

// 
//  n*m の大きさの　array<MazeWallType>　
//  大きさ n * w
// a)現在掘っている場所。
// b)今まで掘っていた場所(分岐のバックトラックのための)の配列(スタックだな) 

open class MazeBoard
{
    let width   : Int
    let height  : Int
    
    var mazeMap : Array<MazeWallType>
    
    public init(width:Int, height:Int)
    {
        // 偶数でないといけないかな？
        
        self.width = width
        self.height = height
        
        // 「掘れる」値で初期化
        var mazeArray = Array<MazeWallType>(repeating: .Digable, count: self.width * self.height)
        
        /*
         // 掘れない壁で囲む         
         ##########
         #********#
         #********#
         #********#
         #********#
         #********#
         ########## 
         */
        
        // 上の壁
        for i in ( 0 ..< width) { mazeArray[i                     ] = .NotDigable }
        
        // 下の壁
        for i in ( 0 ..< width) { mazeArray[i + width * (height-1)] = .NotDigable }
        
        // 左の壁
        for i in ( 0 ..< height){ mazeArray[i * width             ] = .NotDigable }
         
        // 右の壁
        for i in ( 0 ..< height){ mazeArray[i * width + width-1   ] = .NotDigable }

        // スタート位置の設定
        mazeArray[ 1 + width ] = .Start;
        
        // ゴール位置の設定
        mazeArray[ width * (height-1) - 2] = .Goal;
        
        self.mazeMap = mazeArray
    }

    // 現在地を与えられると迷路を掘り進む
    public func doDig(_ pos :MazePos, _ closure: (_ m: MazeBoard ) -> Void = { _ in})
    {
        closure(self)
        
        let theDigables : Array<MazePos> = self.digables(pos)
        
        if theDigables.count == 0 
        {
            return
        }

        // 順列のうちランダムにを選ぶ。
        let thePermutations = theDigables.permutations()
        
        for p in thePermutations[Int(arc4random_uniform(UInt32( thePermutations.count )))]
        {
            // 掘った結果、掘り進めなくなっているかもしれないので、掘れるか否かを確認してから掘る。
            if self.digables(p).count == 0
            {
                continue
            }

            // 掘る
            self.setTypeFor(type:.Digged, pos:p)        
            
            // 掘った先で同じことを繰り返す。いつもの
            doDig(p, closure)
        }
    }
    
    // 上下左右の位置の配列を返す。四方
    public func roundPositions(_ pos:MazePos) -> [MazePos]
    {        
        return [MazePos(0, -1), MazePos(0, 1), MazePos(-1, 0), MazePos(1, 0)].map{ pos + $0 }
    }
        
    // 与えられた地点から四方を見て、掘り進められる方向の配列を返す
    public func digables(_ pos:MazePos) -> [MazePos]
    {
        var result : Array<MazePos> = [] as! Array<MazePos>
        let current = self.typeFor(pos:pos)
        
        if current == .Digable || current == .Digged || current == .Start
        {
            // 掘削可能な位置を返す
            result = self.roundPositions(pos) // 四方の座標を配列で取得
                .flatMap
                {                    
                    // 掘削可能な柔らかさのセルの配列を返す
                    return ( self.typeFor( pos: $0 ) == .Digable || self.typeFor( pos: $0 ) == .Goal ) ? $0 : nil
                } 
                .flatMap
                {
                    let centerPos : MazePos = $0

                    // 掘った場合に、壁を突き抜けて別の通路につながってしまわないかをチェックする。                    
                    let count = (
                        self.roundPositions(centerPos)
                            .filter
                            {       
                                // 現在地は通路でも構わないので除外
                                // Digged(通路)につながっているものを探す
                                ( $0 != pos ) && ( self.typeFor( pos: $0 ) == .Digged )
                                
                            }).count
                    
                    return count == 0 ? centerPos : nil
                }
        }
        
        return result
    }
    
    // 与えられた位置のTypeを変更する
    public func setTypeFor(type:MazeWallType, pos:MazePos)
    {
        self.mazeMap[self.indexFor(pos:pos)] = type
    }
    
    // 与えられた地点のMazeWallTypeを返す
    public func typeFor(pos:MazePos) -> MazeWallType
    {
        return self.mazeMap[self.indexFor(pos:pos)]
    }
    
    
    // 与えられた地点のindexを返す
    func indexFor(pos:MazePos) -> Int
    {
        return pos.x + pos.y * self.width
    }
    
    // MazeWallTypeから文字列の変換
    func mazeWallTypeToString(type :MazeWallType) -> String
    {
        switch type
        {
            case .Digable:    return " *" // 掘れる
            case .NotDigable: return " #" // 掘れない
            case .Digged:     return "  " // 掘ったから通路になった
            case .Start:      return " S" // スタート位置。
            case .Goal:       return " G" // ゴール位置。
        }
    }
    
    public func size() -> (width:Int, height:Int)
    {
        return (self.width, self.height)
    }
    
    public func DebugPrint()
    {        
        print(
            (0 ..< self.height) .map
                {
                    i in
                    
                    (0 ..< self.width).map
                        {
                            j in
                            
                            self.mazeWallTypeToString( type: self.typeFor( pos: MazePos(j, i) ))
                            
                        }.joined() + "\n"
                    
                }.joined()
        )            
    }
}
